# yellowbrick.contrib.statsmodels
# Implements wrappers around hte statsmodels library to use Yellowbrick with.
#
# Author:  Benjamin Bengfort <benjamin@bengfort.com>
# Created: Wed Apr 04 13:13:24 2018 -0400
#
# ID: __init__.py [d6ebc39] benjamin@bengfort.com $

"""
Implements wrappers around hte statsmodels library to use Yellowbrick with.
"""

##########################################################################
## Imports
##########################################################################

from .base import StatsModelsWrapper
